{ $Id: sortgrid.pas 364 2020-11-13 16:33:16Z Heiner $

 *****************************************************************************
 
                               sortgrid.pas
                               ------------
          An interface extending tStringGrid for Delphi or Lazarus.
               Delphi tStringGrid is not capable of sorting.                        .
             Lazarus tStringGrid is already capable of sorting,
                     but natively only by Strings.
               tSortGrid allows sorting by Numbers or DateTime.
                     Initial Revision : May 2011

 *****************************************************************************

 This file is distributed under the Library GNU General Public License
 (see the file COPYING.LGPL) with the following modification:

 As a special exception, the copyright holder of this file gives you
 permission to link this file with independent modules to produce an
 executable, regardless of the license terms of these independent modules,
 and to copy and distribute the resulting executable under terms of your choice,
 provided that you also meet, for each linked independent module, the terms
 and conditions of the license of that module. An independent module is a
 module which is not derived from or based on this file. If you modify this
 file, you may extend this exception to your version of the file, but
 you are not obligated to do so. If you do not wish to do so, delete this
 exception statement from your version.

 In order to receive a copy of the file COPYING.LGPL, contact:
      Free Software Foundation, Inc.,
      675 Mass Ave
      Cambridge, MA  02139
      USA

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 *****************************************************************************

TSortGrid for Lazarus or Delphi
Copyright (C) 2011 Heinrich Wolf.
http://www.Wolf-Fuerth.de/
}

unit sortgrid;

{$ifdef fpc}
  {$mode objfpc}{$H+}
{$endif}

interface

uses Classes,
     SysUtils,
     {$ifdef fpc}
       LResources,
       Windows,
     {$else}
       {$ifdef Ver330}
          UITypes,
       {$endif}
       WinTypes,
     {$endif}
     Graphics,
     Forms,
     Controls,
     Grids,
     wolfutil;

type tMayContinue         = function : Boolean of Object;
     tCompare             = function (a, b : LongInt) : Integer of Object;
     tSwap                = procedure(a, b : LongInt) of Object;
     tFormatType          = (sgfString, sgfText, sgfInteger, sgfFloat,
                             sgfDateTime);
     tGridOrientation     = (oSortByCols, oSortByRows);
     {$ifndef fpc}
       tOnCompareCells    = procedure(Sender: TObject;
                                      ACol, ARow, BCol, BRow: LongInt;
                                      var Result: integer) of object;
       tSortOrder         = (soAscending, soDescending);
     {$else}
       tOSVERSIONINFO     = record
                              dwMajorVersion : DWord;
                            end;
     {$endif}
     tOnColRowExchanged   = procedure(Sender: TObject; IsColumn: Boolean;
                                      sIndex, tIndex: LongInt) of Object;

     tGridSort            = Class(tSort)
       fMayContinue       : tMayContinue;
       fCompare           : tCompare;
       fSwap              : tSwap;
       Constructor          Create(GridMayContinue : tMayContinue;
                                   GridCompare     : tCompare;
                                   GridSwap        : tSwap);
       function             MayContinue : Boolean; override;
       function             Compare(a, b : LongInt) : Integer; override;
       procedure            Swap(a, b : LongInt); override;
     end;

     tSortArray           = Array[0..0] of tSortOrder;
     pSortArray           = ^tSortArray;
     tFormatArray         = Array[0..0] of tFormatType;
     pFormatArray         = ^tFormatArray;
     tAlignArray          = Array[0..0] of tAlignment;
     pAlignArray          = ^tAlignArray;

     tSortList          = class
     private
       fList            : pSortArray;
       fCount           : LongInt;
       function           GetSort(Pos : LongInt) : tSortOrder;
       procedure          SetSort(Pos : LongInt; NewVal : tSortOrder);
       procedure          SetCount(NewCount : LongInt);
     public
       constructor        Create;
       destructor         Destroy; override;
       property           Count : LongInt read fCount write SetCount;
       property           Items[Pos : LongInt] : tSortOrder
                            read GetSort write SetSort; default;
       procedure          Exchange(Index, WithIndex : LongInt);
     end;

     tFormatList        = class
     private
       fList            : pFormatArray;
       fCount           : LongInt;
       function           GetFormat(Pos : LongInt) : tFormatType;
       procedure          SetFormat(Pos : LongInt; NewVal : tFormatType);
       procedure          SetCount(NewCount : LongInt);
     public
       constructor        Create;
       destructor         Destroy; override;
       property           Count : LongInt write SetCount;
       property           Items[Pos : LongInt] : tFormatType
                            read GetFormat write SetFormat; default;
       procedure          Exchange(Index, WithIndex : LongInt);
     end;

     tAlignList         = class
     private
       fList            : pAlignArray;
       fCount           : LongInt;
       function           GetAlign(Pos : LongInt) : tAlignment;
       procedure          SetAlign(Pos : LongInt; NewVal : tAlignment);
       procedure          SetCount(NewCount : LongInt);
     public
       constructor        Create;
       destructor         Destroy; override;
       property           Count : LongInt write SetCount;
       property           Items[Pos : LongInt] : tAlignment
                            read GetAlign write SetAlign; default;
       procedure          Exchange(Index, WithIndex : LongInt);
     end;

     tSortGrid            = Class(tStringGrid)
     private
       fOrientation       : tGridOrientation;
       fSortOrder         : tSortList;
       fFormatType        : tFormatList;
       fFormatString      : tStringList;
       fAlignment         : tAlignList;
       fIntCells          : tLongestIntMatrix;
       fFloatCells        : tExtendedMatrix;
       fSortColRows       : tLongIntList;
       fSort              : tGridSort;
       fMayContinue       : Boolean;
       fShowArrow         : Boolean;
       fIsSorted          : Boolean;
       fIsSorting         : Boolean;
       fOnCompareCells    : tOnCompareCells;
       fOnColRowExchanged : tOnColRowExchanged;
       procedure            CleanUp;
       function             GetOrientation : tGridOrientation;
       procedure            SetOrientation(NewOrientation : tGridOrientation);
       function             GetGridColCount : LongInt;
       procedure            NewGridColCount(OldCount : LongInt;
                                            var NewCount : LongInt);
       procedure            SetGridColCount(Count : LongInt);
       function             GetGridRowCount : LongInt;
       procedure            NewGridRowCount(OldCount : LongInt;
                                            var NewCount : LongInt);
       procedure            SetGridRowCount(Count : LongInt);
       function             GetMayContinue : Boolean;
       function             GetSortOrder(GetSortColRow : LongInt) : TSortOrder;
       procedure            SetSortOrder(NewSortColRow : LongInt;
                                         NewOrder      : TSortOrder);
       function             GetFormatType(GetFormatColRow : LongInt) : tFormatType;
       procedure            SetFormatType(NewFormatColRow : LongInt;
                                          FormatType      : tFormatType);
       function             GetFormatString(GetFormatColRow : LongInt) : String;
       procedure            SetFormatString(NewFormatColRow : LongInt;
                                            FormatString    : String);
       function             GetAlignment(GetAlignColRow : LongInt) : tAlignment;
       procedure            SetAlignment(NewAlignColRow : LongInt;
                                         Alignment   : tAlignment);
       function             GetCell(ACol, ARow : LongInt) : String;
       procedure            SetCell(ACol, ARow : LongInt; Content : String);
       procedure            CheckSortCol(ACol : LongInt);
       procedure            CheckSortRow(ARow : LongInt);
       function             GetSortColRowCount : LongInt;
       function             GetSortColRow(Pos : LongInt) : LongInt;
       procedure            GridRowsDeleted(Index, Count : LongInt);
       procedure            GridColsDeleted(Index, Count : LongInt);
       procedure            GridColsExchanged(Index, WithIndex : LongInt);
       procedure            GridRowsExchanged(Index, WithIndex : LongInt);
       procedure            GridColsInserted(AtCol, Count : LongInt);
       procedure            GridRowsInserted(AtRow, Count : LongInt);
       procedure            GridColMoved(FromIndex, ToIndex : LongInt);
       procedure            GridRowMoved(FromIndex, ToIndex : LongInt);
       procedure            SwapColRows(ColRowA, ColRowB : LongInt);
       procedure            SetUnsorted;
       function             CompareColRows(ColRowA, ColRowB : LongInt) : Integer;
       procedure            CreateHelpers;
     protected
       procedure            DrawCell(ACol, ARow: Longint; ARect: TRect;
                                     AState: TGridDrawState); override;
       procedure            SetEditText(ACol, ARow: Longint;
                                        const Value: String); override;
       {$ifdef fpc}
         function           DoCompareCells(Acol, ARow, Bcol, BRow: LongInt) :
                              Integer; override;
         procedure          ColRowDeleted(IsColumn: Boolean; index: LongInt);
                              override;
         procedure          ColRowExchanged(IsColumn: Boolean;
                                            index, WithIndex: LongInt);
                              override;
         procedure          ColRowInserted(IsColumn: boolean; index: LongInt);
                              override;
         procedure          ColRowMoved(IsColumn: Boolean;
                                        FromIndex, ToIndex: LongInt); override;
       {$else}
         procedure          ColumnMoved(FromIndex, ToIndex : LongInt); override;
         procedure          RowMoved(FromIndex, ToIndex : LongInt); override;
         function           DelphiCompareCells(Acol, ARow, Bcol, BRow: LongInt) :
                              Integer;
         function           DoCompareCells(Acol, ARow, Bcol, BRow: LongInt) :
                              Integer;
         procedure          ColRowExchanged(IsColumn : Boolean;
                                            Index, WithIndex : LongInt);
       {$endif}
     public
       constructor          Create(AOwner: TComponent); override;
       destructor           Destroy; override;
       property             Orientation : tGridOrientation
                              read GetOrientation write SetOrientation;
       property             MayContinue : Boolean
                              read GetMayContinue write fMayContinue;
       property             SortOrder[SColRow : LongInt] : TSortOrder
                              read GetSortOrder write SetSortOrder;
       property             FormatType[FormatColRow : LongInt] : tFormatType
                              read GetFormatType write SetFormatType;
       property             FormatString[FormatColRow : LongInt] : String
                              read GetFormatString write SetFormatString;
       property             Alignment[AlignCol : LongInt] : tAlignment
                              read GetAlignment write SetAlignment;
       property             Cells[ACol, ARow : LongInt] : String
                              read GetCell write SetCell;
       property             IsSorted : Boolean read fIsSorted;
       property             SortColRowCount : LongInt read GetSortColRowCount;
       property             SortColRows[i : LongInt] : LongInt read GetSortColRow;
       function             DefaultCompareCells(
                              Acol, ARow, Bcol, BRow: LongInt) : Integer;
       procedure
                            DoSort(FromColRow, ToColRow : LongInt);
       procedure            DeleteCols(FromCol, ToCol : LongInt);
       procedure            DeleteRows(FromRow, ToRow : LongInt);
       procedure            InsertCols(AtCol,   Count : LongInt);
       procedure            InsertRows(AtRow,   Count : LongInt);
       procedure            ExactWidthCell(ACol, ARow : LongInt);
       procedure            ExactWidthRow(ARow : LongInt);
       procedure            MinWidthCell(ACol, ARow : LongInt);
       procedure            MinWidthRow(ARow : LongInt);
       function             RequiredWidthCol(ACol : LongInt) : LongInt;
       procedure            MinWidthCol(ACol : LongInt);
       procedure            ExactWidthCol(ACol : LongInt);
       procedure            FillWidth(ACol : LongInt);
       procedure            ClearSortColRows;
       procedure            AddSortColRow(NewSortColRow : LongInt);
     published
       property             ColCount : LongInt
                              read GetGridColCount write SetGridColCount;
       property             RowCount : LongInt
                              read GetGridRowCount write SetGridRowCount;
       property             ShowArrow : Boolean
                              read fShowArrow      write fShowArrow;
       {$ifndef fpc}
         property           OnCompareCells : tOnCompareCells
                              read fOnCompareCells write fOnCompareCells;
         property           OnColRowExchanged : tOnColRowExchanged
                              read fOnColRowExchanged write fOnColRowExchanged;
       {$endif}
     end;

var OSVERSIONINFO : tOSVERSIONINFO;

procedure Register;

implementation

const
{$ifdef fpc}
  MinWidth = 8;
{$else}
  MinWidth = 4;
{$endif}

procedure Register;
  begin
    RegisterComponents('Sort Grid', [tSortGrid]);
  end;

constructor tSortList.Create;
  begin
    fList  := Nil;
    fCount := 0;
  end;

destructor tSortList.Destroy;
  begin
    if fList <> Nil then
      FreeMem(fList, fCount * sizeof(TSortOrder));
    fList  := Nil;
    fCount := 0;
  end;

function tSortList.GetSort(Pos : LongInt) : TSortOrder;
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tSortList.GetSort(' + IntToStr(Pos)
                             + ') out of range')
    else
      {$R-}
        GetSort := fList^[Pos];
      {$R+}
  end;

procedure tSortList.SetSort(Pos : LongInt; NewVal : TSortOrder);
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tSortList.SetSort(' + IntToStr(Pos)
                             + ') out of range')
    else
      {$R-}
        fList^[Pos] := NewVal;
      {$R+}
  end;

procedure tSortList.SetCount(NewCount : LongInt);
  var     NewList : pSortArray;
          i, j    : LongInt;
  begin
    if NewCount < 0 then
      Raise Exception.Create('invalid tSortList.SetCount('
                             + IntToStr(NewCount) + ')')
    else
      if NewCount <> fCount then
        begin
          GetMem(NewList, NewCount * sizeof(TSortOrder));
          if NewList = Nil then
            Raise Exception.Create('out of memory in tSortList.SetCount')
          else
            begin
              if NewCount > fCount then
                j := fCount   - 1
              else
                j := NewCount - 1;
              {$R-}
                for i := 0 to j do
                  NewList^[i] := fList^[i];
                for i := fCount to NewCount - 1 do
                  NewList^[i] := soAscending;
              {$R+}
              if fList <> Nil then
                FreeMem(fList, fCount * sizeof(TSortOrder));
              fList  := NewList;
              fCount := NewCount;
            end;
        end;
  end;

procedure tSortList.Exchange(Index, WithIndex : LongInt);
  var     Tmp : TSortOrder;
  begin
    if (Index < 0) or (Index >= fCount) then
      Raise Exception.Create('tSortList.Exchange(Index '
                             + IntToStr(Index) + ' out of range)')
    else
      if (WithIndex < 0) or (WithIndex >= fCount) then
        Raise Exception.Create('tSortList.Exchange(WithIndex '
                               + IntToStr(WithIndex) + ' out of range)')
      else
        begin
          {$R-}
            Tmp               := fList^[Index];
            fList^[Index]     := fList^[WithIndex];
            fList^[WithIndex] := Tmp;
          {$R+}
        end;
  end;

constructor tFormatList.Create;
  begin
    fList  := Nil;
    fCount := 0;
  end;

destructor tFormatList.Destroy;
  begin
    if fList <> Nil then
      FreeMem(fList, fCount * sizeof(tFormatType));
    fList  := Nil;
    fCount := 0;
  end;

function tFormatList.GetFormat(Pos : LongInt) : tFormatType;
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tFormatList.GetFormat(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        GetFormat := fList^[Pos];
      {$R+}
  end;

procedure tFormatList.SetFormat(Pos : LongInt; NewVal : tFormatType);
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tFormatList.SetFormat(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        fList^[Pos] := NewVal;
      {$R+}
  end;

procedure tFormatList.SetCount(NewCount : LongInt);
  var     NewList : pFormatArray;
          i, j    : LongInt;
  begin
    if NewCount < 0 then
      Raise Exception.Create('invalid tFormatList.SetCount('
                             + IntToStr(NewCount) + ')')
    else
      if NewCount <> fCount then
        begin
          GetMem(NewList, NewCount * sizeof(tFormatType));
          if NewList = Nil then
            Raise Exception.Create('out of memory in tFormatList.SetCount')
          else
            begin
              if NewCount > fCount then
                j := fCount   - 1
              else
                j := NewCount - 1;
              {$R-}
                for i := 0 to j do
                  NewList^[i] := fList^[i];
                for i := fCount to NewCount - 1 do
                  NewList^[i] := sgfText;
              {$R+}
              if fList <> Nil then
                FreeMem(fList, fCount * sizeof(tFormatType));
              fList  := NewList;
              fCount := NewCount;
            end;
        end;
  end;

procedure tFormatList.Exchange(Index, WithIndex : LongInt);
  var     Tmp : tFormatType;
  begin
    if (Index < 0) or (Index >= fCount) then
      Raise Exception.Create('tFormatList.Exchange(Index '
                             + IntToStr(Index) + ' out of range)')
    else
      if (WithIndex < 0) or (WithIndex >= fCount) then
        Raise Exception.Create('tFormatList.Exchange(WithIndex '
                               + IntToStr(WithIndex) + ' out of range)')
      else
        begin
          {$R-}
            Tmp               := fList^[Index];
            fList^[Index]     := fList^[WithIndex];
            fList^[WithIndex] := Tmp;
          {$R+}
        end;
  end;

constructor tAlignList.Create;
  begin
    fList  := Nil;
    fCount := 0;
  end;

destructor tAlignList.Destroy;
  begin
    if fList <> Nil then
      FreeMem(fList, fCount * sizeof(tAlignment));
    fList  := Nil;
    fCount := 0;
  end;

function tAlignList.GetAlign(Pos : LongInt) : tAlignment;
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tAlignList.GetAlign(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        GetAlign := fList^[Pos];
      {$R+}
  end;

procedure tAlignList.SetAlign(Pos : LongInt; NewVal : tAlignment);
  begin
    if (Pos < 0) or (Pos >= fCount) then
      Raise Exception.Create('tAlignList.SetAlign(' + IntToStr(Pos)
                             + ' out of range)')
    else
      {$R-}
        fList^[Pos] := NewVal;
      {$R+}
  end;

procedure tAlignList.SetCount(NewCount : LongInt);
  var     NewList : pAlignArray;
          i, j    : LongInt;
  begin
    if NewCount < 0 then
      Raise Exception.Create('invalid tAlignList.SetCount('
                             + IntToStr(NewCount) + ')')
    else
      if NewCount <> fCount then
        begin
          GetMem(NewList, NewCount * sizeof(tAlignment));
          if NewList = Nil then
            Raise Exception.Create('out of memory in tAlignList.SetCount')
          else
            begin
              if NewCount > fCount then
                j := fCount   - 1
              else
                j := NewCount - 1;
              {$R-}
                for i := 0 to j do
                  NewList^[i] := fList^[i];
                for i := fCount to NewCount - 1 do
                  NewList^[i] := taLeftJustify;
              {$R+}
              if fList <> Nil then
                FreeMem(fList, fCount * sizeof(tAlignment));
              fList  := NewList;
              fCount := NewCount;
            end;
        end;
  end;

procedure tAlignList.Exchange(Index, WithIndex : LongInt);
  var     Tmp : tAlignment;
  begin
    if (Index < 0) or (Index >= fCount) then
      Raise Exception.Create('tAlignList.Exchange(Index '
                             + IntToStr(Index) + ' out of range)')
    else
      if (WithIndex < 0) or (WithIndex >= fCount) then
        Raise Exception.Create('tAlignList.Exchange(WithIndex '
                               + IntToStr(WithIndex) + ' out of range)')
      else
        begin
          {$R-}
            Tmp               := fList^[Index];
            fList^[Index]     := fList^[WithIndex];
            fList^[WithIndex] := Tmp;
          {$R+}
        end;
  end;

Constructor tGridSort.Create(GridMayContinue : tMayContinue;
                             GridCompare     : tCompare;
                             GridSwap        : tSwap);
  begin
    fMayContinue := GridMayContinue;
    fCompare     := GridCompare;
    fSwap        := GridSwap;
  end;

function tGridSort.MayContinue : Boolean;
  begin
    {$ifdef fpc}
      MayContinue := fMayContinue();
    {$else}
      MayContinue := fMayContinue;
    {$endif}
  end;

function tGridSort.Compare(a, b : LongInt) : Integer;
  begin
    Compare := fCompare(a, b);
  end;

procedure tGridSort.Swap(a, b : LongInt);
  begin
    fSwap(a, b);
  end;

constructor tSortGrid.Create(AOwner: TComponent);
  begin
    inherited Create(AOwner);
    fOnCompareCells    := Nil;
    fOnColRowExchanged := Nil;
    fOrientation       := oSortByCols;
    fIsSorted          := False;
    fIsSorting         := False;
    fMayContinue       := True;
    fShowArrow         := True;
    try
      fSortOrder       := tSortList.Create;
    except
      fSortOrder       := Nil;
    end;
    try
      fFormatType      := tFormatList.Create;
    except
      fFormatType      := Nil;
    end;
    try
      fFormatString    := tStringList.Create;
    except
      fFormatString    := Nil;
    end;
    try
      fAlignment       := tAlignList.Create;
    except
      fAlignment       := Nil;
    end;
    try
      fIntCells        := tLongestIntMatrix.Create;
    except
      fIntCells        := Nil;
    end;
    try
      fFloatCells      := tExtendedMatrix.Create;
    except
      fFloatCells      := Nil;
    end;
    try
      fSortColRows     := tLongIntList.Create;
    except
      fSortColRows     := Nil;
    end;
    try
      {$ifdef fpc}
        fSort          := tGridSort.Create(@GetMayContinue,
                                           @CompareColRows,
                                           @SwapColRows);
      {$else}
        fSort          := tGridSort.Create(GetMayContinue,
                                           CompareColRows,
                                           SwapColRows);
      {$endif}
    except
      fSort            := Nil;
    end;
    if   (fSortOrder    = Nil)
      or (fFormatType   = Nil)
      or (fFormatString = Nil)
      or (fAlignment    = Nil)
      or (fIntCells     = Nil)
      or (fFloatCells   = Nil)
      or (fSortColRows  = Nil)
      or (fSort         = Nil)
    then
      begin
        CleanUp;
        raise Exception.Create('error in tSortGrid.Create');
      end;
    fAlignment.Count := inherited ColCount;
    CreateHelpers;
  end;

destructor tSortGrid.Destroy;
  begin
    CleanUp;
    inherited Destroy;
  end;

procedure tSortGrid.CreateHelpers;
  var i : LongInt;
  begin
    case Orientation of
      oSortByCols:
        begin
          fSortOrder.Count  := inherited ColCount;
          fFormatType.Count := inherited ColCount;
          for i := 0 to inherited ColCount - 1 do
            fFormatString.Add('');
        end;
      oSortByRows:
        begin
          fSortOrder.Count  := inherited RowCount;
          fFormatType.Count := inherited RowCount;
          for i := 0 to inherited RowCount - 1 do
            fFormatString.Add('');
        end;
    end;
  end;

procedure tSortGrid.CleanUp;
  begin
    fSortOrder.Free;
    fFormatType.Free;
    fFormatString.Free;
    fAlignment.Free;
    fIntCells.Free;
    fFloatCells.Free;
    fSortColRows.Free;
    fSort.Free;
    fOnCompareCells    := Nil;
    fOnColRowExchanged := Nil;
    fIsSorted          := False;
    fIsSorting         := False;
    fMayContinue       := True;
    fShowArrow         := True;
    fSortOrder         := Nil;
    fFormatType        := Nil;
    fFormatString      := Nil;
    fAlignment         := Nil;
    fIntCells          := Nil;
    fFloatCells        := Nil;
    fSortColRows       := Nil;
    fSort              := Nil;
  end;

procedure tSortGrid.SetUnsorted;
  begin
    if fIsSorted then
      begin
        fIsSorted := False;
        if fShowArrow then
          if fSortColRows.Count > 0 then
            case Orientation of
              oSortByCols:
                InvalidateCell(fSortColRows[0], 0);
              oSortByRows:
                InvalidateCell(0, fSortColRows[0]);
            end;
      end;
  end;

procedure tSortGrid.SwapColRows(ColRowA, ColRowB : LongInt);
  begin
    if fOrientation = oSortByCols then
      ColRowExchanged(False, ColRowA, ColRowB)
    else
      ColRowExchanged(True, ColRowA, ColRowB);
  end;

function tSortGrid.DefaultCompareCells(
           Acol, ARow, Bcol, BRow: LongInt) : Integer;
  begin
    DefaultCompareCells := AnsiCompareText(inherited Cells[ACol, ARow],
                                           inherited Cells[BCol, BRow]);
  end;

{$ifndef fpc}
  function tSortGrid.DelphiCompareCells(
             Acol, ARow, Bcol, BRow: LongInt) : Integer;
    var    i : Integer;
    begin
      if Assigned(fOnCompareCells) then
        begin
          fOnCompareCells(Self, ACol, ARow, BCol, BRow, i);
          DelphiCompareCells := i;
        end
      else
        DelphiCompareCells := AnsiCompareText(inherited Cells[ACol, ARow],
                                              inherited Cells[BCol, BRow]);
    end;
{$endif}

function tSortGrid.DoCompareCells(ACol, ARow, BCol, BRow: LongInt) : Integer;
  var    i,          Ascending  : Integer;
         IntACell,   IntBCell   : LongestInt;
         FloatACell, FloatBCell : Extended;
         StrACell,   StrBCell   : String;
  begin
    if Orientation = oSortByCols then
      begin
        if ACol <> BCol then
          {$ifdef fpc}
            DoCompareCells := inherited DoCompareCells(ACol, ARow, BCol, BRow)
          {$else}
            DoCompareCells := DelphiCompareCells(ACol, ARow, BCol, BRow)
          {$endif}
        else
          begin
            case fSortOrder[ACol] of
              soAscending : Ascending := 1
            else
              Ascending := -1;
            end;
            case fFormatType[ACol] of
              sgfInteger :
                begin
                  IntACell := fIntCells[ACol, ARow];
                  IntBCell := fIntCells[BCol, BRow];
                  if IntACell = IntBCell then
                    DoCompareCells := 0
                  else
                    if IntACell > IntBCell then
                      DoCompareCells :=  Ascending
                    else
                      DoCompareCells := -Ascending;
                end;
              sgfFloat,
              sgfDateTime :
                begin
                  FloatACell := fFloatCells[ACol, ARow];
                  FloatBCell := fFloatCells[BCol, BRow];
                  if FloatACell = FloatBCell then
                    DoCompareCells := 0
                  else
                    if FloatACell > FloatBCell then
                      DoCompareCells :=  Ascending
                    else
                      DoCompareCells := -Ascending;
                end;
              sgfText :
                begin
                  {$ifdef fpc}
                    i := inherited DoCompareCells(ACol, ARow, BCol, BRow);
                  {$else}
                    i := DelphiCompareCells(ACol, ARow, BCol, BRow);
                  {$endif}
                  if i = 0 then
                    DoCompareCells := 0
                  else
                    if i > 0 then
                      DoCompareCells :=  Ascending
                    else
                      DoCompareCells := -Ascending;
                end
            else
              StrACell := inherited Cells[ACol, ARow];
              StrBCell := inherited Cells[BCol, BRow];
              if StrACell = StrBCell then
                DoCompareCells := 0
              else
                if StrACell > StrBCell then
                  DoCompareCells :=  Ascending
                else
                  DoCompareCells := -Ascending;
            end;
          end;
      end
    else
      begin
        if ARow <> BRow then
          {$ifdef fpc}
            DoCompareCells := inherited DoCompareCells(ACol, ARow, BCol, BRow)
          {$else}
            DoCompareCells := DelphiCompareCells(ACol, ARow, BCol, BRow)
          {$endif}
        else
          begin
            case fSortOrder[ARow] of
              soAscending : Ascending := 1
            else
              Ascending := -1;
            end;
            case fFormatType[ARow] of
              sgfInteger :
                begin
                  IntACell := fIntCells[ARow, ACol];
                  IntBCell := fIntCells[BRow, BCol];
                  if IntACell = IntBCell then
                    DoCompareCells := 0
                  else
                    if IntACell > IntBCell then
                      DoCompareCells :=  Ascending
                    else
                      DoCompareCells := -Ascending;
                end;
              sgfFloat,
              sgfDateTime :
                begin
                  FloatACell := fFloatCells[ARow, ACol];
                  FloatBCell := fFloatCells[BRow, BCol];
                  if FloatACell = FloatBCell then
                    DoCompareCells := 0
                  else
                    if FloatACell > FloatBCell then
                      DoCompareCells :=  Ascending
                    else
                      DoCompareCells := -Ascending;
                end;
              sgfText :
                begin
                  {$ifdef fpc}
                    i := inherited DoCompareCells(ACol, ARow, BCol, BRow);
                  {$else}
                    i := DelphiCompareCells(ACol, ARow, BCol, BRow);
                  {$endif}
                  if i = 0 then
                    DoCompareCells := 0
                  else
                    if i > 0 then
                      DoCompareCells :=  Ascending
                    else
                      DoCompareCells := -Ascending;
                end
            else
              StrACell := inherited Cells[ACol, ARow];
              StrBCell := inherited Cells[BCol, BRow];
              if StrACell = StrBCell then
                DoCompareCells := 0
              else
                if StrACell > StrBCell then
                  DoCompareCells :=  Ascending
                else
                  DoCompareCells := -Ascending;
            end;
          end;
      end;
  end;

function tSortGrid.CompareColRows(ColRowA, ColRowB : LongInt) : Integer;
  var    c : LongInt;
         i : Integer;
  begin
    i := 0;
    c := 0;
    while (i = 0) and (c < fSortColRows.Count) do
      begin
        if fOrientation = oSortByCols then
          i := DoCompareCells(fSortColRows[c], ColRowA, fSortColRows[c], ColRowB)
        else
          i := DoCompareCells(ColRowA, fSortColRows[c], ColRowB, fSortColRows[c]);
        inc(c);
      end;
    CompareColRows := i;
  end;

function tSortGrid.GetMayContinue : Boolean;
  begin
    Application.ProcessMessages;
    GetMayContinue := fMayContinue;
  end;

function tSortGrid.GetSortOrder(GetSortColRow : LongInt) : TSortOrder;
  begin
    GetSortOrder := fSortOrder[GetSortColRow];
  end;

procedure tSortGrid.SetSortOrder(NewSortColRow : LongInt;
                                 NewOrder      : TSortOrder);
  var     c : LongInt;
  begin
    if fSortOrder[NewSortColRow] <> NewOrder then
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.SetSortOrder while sorting')
      else
        begin
          for c := 0 to fSortColRows.Count - 1 do
            if NewOrder = fSortOrder[c] then
              SetUnsorted;
          fSortOrder[NewSortColRow] := NewOrder;
        end;
  end;

function tSortGrid.GetFormatType(GetFormatColRow : LongInt) : tFormatType;
  begin
    GetFormatType := fFormatType[GetFormatColRow];
  end;

procedure tSortGrid.SetFormatType(NewFormatColRow : LongInt;
                                  FormatType      : tFormatType);
  var     c : LongInt;
  begin
    if fFormatType[NewFormatColRow] <> FormatType then
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.SetFormatType while sorting')
      else
        begin
          for c := 0 to fSortColRows.Count - 1 do
            if NewFormatColRow = fSortColRows[c] then
              SetUnsorted;
          fFormatType[NewFormatColRow] := FormatType;
        end;
  end;

function tSortGrid.GetAlignment(GetAlignColRow : LongInt) : tAlignment;
  begin
    GetAlignment := fAlignment[GetAlignColRow];
  end;

procedure tSortGrid.SetAlignment(NewAlignColRow : LongInt;
                                 Alignment      : tAlignment);
  begin
    fAlignment[NewAlignColRow] := Alignment;
    {$ifdef Ver80}
      Invalidate;
    {$else}
      if fOrientation = oSortByCols then
        InvalidateCol(NewAlignColRow)
      else
        InvalidateRow(NewAlignColRow)
    {$endif}
  end;

function tSortGrid.GetFormatString(GetFormatColRow : LongInt) : String;
  begin
    GetFormatString := fFormatString[GetFormatColRow];
  end;

procedure tSortGrid.SetFormatString(NewFormatColRow : LongInt;
                                    FormatString    : String);
  var     c : LongInt;
  begin
    if fFormatString[NewFormatColRow] <> FormatString then
      begin
        for c := 0 to fSortColRows.Count - 1 do
          if fSortColRows[c] = NewFormatColRow then
            if fFormatType[fSortColRows[c]] = sgfDateTime then
              SetUnsorted;
        fFormatString[NewFormatColRow] := FormatString;
      end;
  end;

function tSortGrid.GetSortColRow(Pos : LongInt) : LongInt;
  begin
    if (Pos < 0) or (Pos >= fSortColRows.Count) then
      Raise Exception.Create('tSortList.SortSortColRows(' + IntToStr(Pos)
                             + ') out of range')
    else
      {$R-}
        GetSortColRow := fSortColRows[Pos];
      {$R-}
  end;

procedure tSortGrid.SetCell(ACol, ARow : LongInt; Content : String);
  begin
    inherited Cells[ACol, ARow] := Content;
    CheckSortCol(ACol);
    CheckSortRow(ARow);
  end;

function tSortGrid.GetCell(ACol, ARow : LongInt) : String;
  begin
    GetCell := inherited Cells[ACol, ARow];
  end;

procedure tSortGrid.CheckSortCol(ACol : LongInt);
  var     c : LongInt;
  begin
    if fOrientation = oSortByCols then
      for c := 0 to fSortColRows.Count - 1 do
        if fSortColRows[c] = ACol then
          SetUnsorted;
  end;

procedure tSortGrid.CheckSortRow(ARow : LongInt);
  var     c : LongInt;
  begin
    if fOrientation = oSortByRows then
      for c := 0 to fSortColRows.Count - 1 do
        if fSortColRows[c] = ARow then
          SetUnsorted;
  end;

function tSortGrid.GetSortColRowCount : LongInt;
  begin
    GetSortColRowCount := fSortColRows.Count;
  end;

procedure tSortGrid.AddSortColRow(NewSortColRow : LongInt);
  begin
    if (NewSortColRow < 0)
       or ((NewSortColRow >= inherited ColCount) and (fOrientation = oSortByCols))
       or ((NewSortColRow >= inherited RowCount) and (fOrientation = oSortByRows))
    then
      raise Exception.Create(
        'tSortGrid.AddSortCol(' + IntToStr(NewSortColRow) + ' out of range)')
    else
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.AddSortColRow while sorting')
      else
        begin
          SetUnsorted;
          fSortColRows.Add(NewSortColRow);
        end;
  end;

procedure tSortGrid.ClearSortColRows;
  begin
    SetUnsorted;
    fSortColRows.Clear;
  end;

function tSortGrid.GetOrientation : tGridOrientation;
  begin
    GetOrientation := fOrientation;
  end;

procedure tSortGrid.SetOrientation(NewOrientation : tGridOrientation);
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.SetOrientation while sorting')
    else
      begin
        fOrientation := NewOrientation;
        SetUnsorted;
        fSortColRows.Clear;
      end;
  end;

function tSortGrid.GetGridColCount : LongInt;
  begin
    GetGridColCount := inherited ColCount;
  end;

procedure tSortGrid.NewGridColCount(OldCount : LongInt; var NewCount : LongInt);
  var     i : LongInt;
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.SetGridColCount while sorting')
    else
      begin
        {$ifndef fpc}
          { Delphi TStringGrid does not allow RowCount 0 }
          if NewCount <= 0 then
            NewCount := 1;
        {$endif}
        if fOrientation = oSortByCols then
          begin
            fAlignment.Count  := NewCount;
            fSortOrder.Count  := NewCount;
            fFormatType.Count := NewCount;
            for i := OldCount to NewCount - 1 do
              fFormatString.Add('');
            for i := OldCount - 1 downto NewCount do
              fFormatString.Delete(i);
            i := fSortColRows.Count - 1;
            while i >= 0 do
              begin
                if fSortColRows[i] >= NewCount then
                  fSortColRows.Delete(i);
                dec(i);
              end;
          end;
      end;
  end;

procedure tSortGrid.SetGridColCount(Count : LongInt);
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.SetGridColCount while sorting')
    else
      begin
        NewGridColCount(inherited ColCount, Count);
        inherited ColCount := Count;
      end;
  end;

function tSortGrid.GetGridRowCount : LongInt;
  begin
    GetGridRowCount := inherited RowCount;
  end;

procedure tSortGrid.NewGridRowCount(OldCount : LongInt; var NewCount : LongInt);
  var     i : LongInt;
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.SetGridRowCount while sorting')
    else
      begin
        {$ifndef fpc}
          { Delphi TStringGrid does not allow RowCount 0 }
          if NewCount <= 0 then
            NewCount := 1;
        {$endif}
        if fOrientation = oSortByRows then
          begin
            fSortOrder.Count  := NewCount;
            fFormatType.Count := NewCount;
            fAlignment.Count  := NewCount;
            for i := OldCount to NewCount - 1 do
              fFormatString.Add('');
            for i := OldCount - 1 downto NewCount do
              fFormatString.Delete(i);
            i := fSortColRows.Count - 1;
            while i >= 0 do
              begin
                if fSortColRows[i] >= NewCount then
                  fSortColRows.Delete(i);
                dec(i);
              end;
          end;
      end;
  end;

procedure tSortGrid.SetGridRowCount(Count : LongInt);
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.SetGridRowCount while sorting')
    else
      begin
        NewGridRowCount(inherited RowCount, Count);
        inherited RowCount := Count;
      end;
  end;

procedure tSortGrid.DrawCell(ACol, ARow: Longint; ARect: TRect;
                             AState: TGridDrawState);
  const   HalfArrow    =  3;
          ArrowLength  = 12;
  var     p            : Array[0 .. 2] of tPoint;
          AnAlign      : tAlignment;
          ArrowRect    : tRect;
          TextWidth,
          TextLeft,
          ArrowLeft,
          HorzMargin,
          VertMargin   : LongInt;
          ArrowVisible : Boolean;
  begin
    if fSortColRows.Count = 0 then
      begin
        ArrowVisible := False;
        HorzMargin   := 0;
        VertMargin   := 0;
      end
    else
      if fOrientation = oSortByCols then
        begin
          HorzMargin   := 5;
          VertMargin   := (ARect.Bottom - ARect.Top - ArrowLength) div 2;
          ArrowVisible :=
            (aCol = fSortColRows[0]) and (aRow = 0)
            and (ARect.Left + 2 * (HorzMargin + HalfArrow) < GridWidth)
            and fShowArrow and fIsSorted
        end
      else
        begin
          HorzMargin   := 3;
          VertMargin   := (ARect.Bottom - ARect.Top - 4 * HalfArrow) div 2;
          ArrowVisible :=
            (aRow = fSortColRows[0]) and (aCol = 0)
            and (ARect.Left + 2 * HorzMargin + ArrowLength < GridWidth)
            and fShowArrow and fIsSorted;
        end;
    if fOrientation = oSortByCols then
      AnAlign := Alignment[ACol]
    else
      AnAlign := Alignment[ARow];
    if not ArrowVisible and (AnAlign = taLeftJustify) then
      begin
        inherited DrawCell(ACol, ARow, ARect, AState);
        if OSVERSIONINFO.dwMajorVersion > 5 then
          begin
            //Horizontale Linien ohne Fixed Spalten
            if ARow - FixedRows >= 0  Then
              with Canvas do
                begin
                  Pen.Color := clSilver;
                  Pen.Style := psSolid;
                  MoveTo(ARect.Left  - 1, ARect.Bottom);
                  LineTo(ARect.Right - 1, ARect.Bottom);
                end;
            //Vertikale Linien ohne Fixed Zeilen
            if ACol - FixedCols >= 0 Then
              with Canvas do
                begin
                  Pen.Color := clSilver;
                  Pen.Style := psSolid;
                  MoveTo(ARect.Right, ARect.Top    - 1);
                  LineTo(ARect.Right, ARect.Bottom - 1);
                end;
          end;
      end
    else
      begin
        if gdFixed in AState then
          Canvas.Brush.Color := clBtnFace
        else
          if (AState * [gdSelected, gdFocused] = [gdSelected, gdFocused])
             and (goDrawFocusSelected in Options) or
             (AState = [gdSelected]) 
          then
            Canvas.Brush.Color := clHighlight
          else
            Canvas.Brush.Color := clWindow;
        {$ifdef fpc}
          Canvas.Pen.Color := clWindow;
        {$else}
          Canvas.Pen.Color := Canvas.Brush.Color;
        {$endif}
        Canvas.Rectangle(ARect.Left,  ARect.Top,
                         ARect.Right, ARect.Bottom);
        if gdFixed in AState then
          begin
            Canvas.Pen.Color := clBtnHighlight;
            Canvas.MoveTo(ARect.Left,      ARect.Bottom - 1);
            Canvas.LineTo(ARect.Left,      ARect.Top);
            Canvas.LineTo(ARect.Right - 1, ARect.Top);
            {$ifdef fpc}
              Canvas.Pen.Color := cl3DDKShadow;
            {$else}
              Canvas.Pen.Color := clBtnShadow;
            {$endif}
            Canvas.LineTo(ARect.Right - 1, ARect.Bottom - 1);
            Canvas.LineTo(ARect.Left,      ARect.Bottom - 1);
          end
        else
          begin
            {$ifdef fpc}
              Canvas.Pen.Color := clSilver;
              Canvas.MoveTo(ARect.Right - 1, ARect.Top);
              Canvas.LineTo(ARect.Right - 1, ARect.Bottom - 1);
              Canvas.LineTo(ARect.Left,      ARect.Bottom - 1);
            {$endif}
            if OSVERSIONINFO.dwMajorVersion > 5 then
              begin
                //Horizontale Linien ohne Fixed Spalten
                if ARow - FixedRows >= 0  Then
                  with Canvas do
                    begin
                      Pen.Color := clSilver;
                      Pen.Style := psSolid;
                      MoveTo(ARect.Left  - 1, ARect.Bottom);
                      LineTo(ARect.Right - 1, ARect.Bottom);
                    end;
                //Vertikale Linien ohne Fixed Zeilen
                if ACol - FixedCols >= 0 Then
                  with Canvas do
                    begin
                      Pen.Color := clSilver;
                      Pen.Style := psSolid;
                      MoveTo(ARect.Right, ARect.Top    - 1);
                      LineTo(ARect.Right, ARect.Bottom - 1);
                    end;
              end;
          end;
        TextWidth := Canvas.TextWidth(inherited Cells[aCol, aRow]);
        if fOrientation = oSortByCols then
          AnAlign := Alignment[ACol]
        else
          AnAlign := Alignment[ARow];
        case AnAlign of
          taCenter:
            if ArrowVisible then
              begin
                TextLeft := ARect.Left +
                  (ARect.Right - ARect.Left - TextWidth
                   - 2 * (HorzMargin + HalfArrow)) div 2;
                if TextLeft >= ARect.Left + 1 then
                  ArrowLeft := TextLeft + TextWidth
                else
                  begin
                    ArrowLeft := ARect.Right - 2 * (HorzMargin + HalfArrow);
                    TextLeft  := ARect.Left +
                      (ARect.Right - ARect.Left - TextWidth) div 2;
                  end;
              end
            else
              begin
                TextLeft :=
                  ARect.Left + (ARect.Right - ARect.Left - TextWidth) div 2;
                ArrowLeft := TextLeft + TextWidth;
              end;
          taRightJustify:
            begin
              TextLeft  := ARect.Right - 3 - TextWidth;
              ArrowLeft := TextLeft - 2 * HorzMargin - ArrowLength;
              if ArrowLeft < ARect.Left + 1 then
                ArrowLeft := ARect.Left + 1;
            end;
        else
          TextLeft  := ARect.Left + 3;
          ArrowLeft := TextLeft   + TextWidth;
        end;
        {$ifdef fpc}
          Canvas.TextRect(ARect,
                          TextLeft,
                          ARect.Top + (ARect.Bottom - ARect.Top
                                       - Canvas.TextHeight(
                                           inherited Cells[aCol, aRow]))
                                      div 2,
                          inherited Cells[aCol, aRow]);
        {$else}
          Canvas.TextRect(ARect,
                          TextLeft,
                          ARect.Top + 2,
                          inherited Cells[aCol, aRow]);
        {$endif}
        if ArrowVisible then
          begin
            if Orientation = oSortByCols then
              begin
                if ArrowLeft > ARect.Right - 2 * (HorzMargin + HalfArrow) then
                  ArrowLeft := ARect.Right - 2 * (HorzMargin + HalfArrow);
                p[0].x := ArrowLeft + HorzMargin;
                p[2].x := p[0].x    + HalfArrow;
                p[1].x := p[2].x    + HalfArrow;
                ArrowRect.Top    := ARect.Top;
                ArrowRect.Bottom := ARect.Bottom - 1;
                ArrowRect.Left   := ArrowLeft;
                ArrowRect.Right  := ArrowLeft + 2 * (HorzMargin + HalfArrow);
                case fSortOrder[fSortColRows[0]] of
                  soAscending :
                    begin
                      p[0].y := ARect.Bottom - VertMargin;
                      p[2].y := ARect.Top    + VertMargin;
                    end;
                else
                  begin
                    p[0].y := ARect.Top    + VertMargin;
                    p[2].y := ARect.Bottom - VertMargin;
                  end;
                end;
                p[1].y := p[0].y;
              end
            else
              begin
                if ArrowLeft > ARect.Right - 2 * HorzMargin - ArrowLength then
                  ArrowLeft := ARect.Right - 2 * HorzMargin - ArrowLength;
                p[0].y := VertMargin + HalfArrow + ARect.Top;
                p[2].y := p[0].y     + HalfArrow;
                p[1].y := p[2].y     + HalfArrow;
                ArrowRect.Top    := ARect.Top    + VertMargin;
                ArrowRect.Bottom := ARect.Bottom - VertMargin;
                ArrowRect.Left   := ArrowLeft;
                ArrowRect.Right  := ArrowLeft + 2 * HorzMargin + ArrowLength;
                case fSortOrder[fSortColRows[0]] of
                  soAscending :
                    begin
                      p[0].x := ArrowLeft + HorzMargin + ArrowLength;
                      p[2].x := ArrowLeft + HorzMargin;
                    end;
                else
                  begin
                    p[0].x := ArrowLeft + HorzMargin;
                    p[2].x := ArrowLeft + HorzMargin + ArrowLength;
                  end;
                end;
                p[1].x := p[0].x;
              end;
            Canvas.Brush.Color := clBtnFace;
            Canvas.FillRect(ArrowRect);
            Canvas.Pen.Color   := clBlack;
            Canvas.Brush.Color := clBlack;
            Canvas.Polygon(p);
          end;
      end;
  end;

procedure tSortGrid.GridRowsDeleted(Index, Count : LongInt);
  var     i : LongInt;
  begin
    {$ifdef fpc}
      for i := Index to inherited RowCount - 1 do
    {$else}
      for i := Index to inherited RowCount - 1 - Count do
    {$endif}
        if fOrientation = oSortByRows then
          begin
            fAlignment[i]    := fAlignment[i + Count];
            fSortOrder[i]    := fSortOrder[i + Count];
            fFormatType[i]   := fFormatType[i + Count];
            fFormatString[i] := fFormatString[i + Count];
          end;
    if fOrientation = oSortByRows then
      begin
        i := fSortColRows.Count - 1;
        while i >= 0 do
          begin
            if fSortColRows[i] >= Index then
              if fSortColRows[i] >= Index + Count then
                fSortColRows[i] := fSortColRows[i] - Count
              else
                begin
                  fSortColRows.Delete(i);
                  SetUnsorted;
                end;
            dec(i);
          end;
      end;
  end;

procedure tSortGrid.GridColsDeleted(Index, Count : LongInt);
  var     i : LongInt;
  begin
    {$ifdef fpc}
      for i := Index to inherited ColCount - 1 do
    {$else}
      for i := Index to inherited ColCount - 1 - Count do
    {$endif}
        if fOrientation = oSortByCols then
          begin
            fAlignment[i]    := fAlignment[i + Count];
            fSortOrder[i]    := fSortOrder[i + Count];
            fFormatType[i]   := fFormatType[i + Count];
            fFormatString[i] := fFormatString[i + Count];
          end;
    if fOrientation = oSortByCols then
      begin
        i := fSortColRows.Count - 1;
        while i >= 0 do
          begin
            if fSortColRows[i] >= Index then
              if fSortColRows[i] >= Index + Count then
                fSortColRows[i] := fSortColRows[i] - Count
              else
                begin
                  fSortColRows.Delete(i);
                  SetUnsorted;
                end;
            dec(i);
          end;
      end;
  end;

procedure tSortGrid.GridColsExchanged(Index, WithIndex : LongInt);
  var     i : LongInt;
  begin
    if fOrientation = oSortByCols then
      begin
        fAlignment.Exchange(Index,    WithIndex);
        fSortOrder.Exchange(Index,    WithIndex);
        fFormatType.Exchange(Index,   WithIndex);
        fFormatString.Exchange(Index, WithIndex);
        for i := 0 to fSortColRows.Count - 1 do
          if fSortColRows[i] = Index then
            fSortColRows[i] := WithIndex
          else
            if fSortColRows[i] = WithIndex then
              fSortColRows[i] := Index;
      end
    else
      begin
        fIntCells.Exchange(Index,   WithIndex);
        fFloatCells.Exchange(Index, WithIndex);
      end;
  end;

procedure tSortGrid.GridRowsExchanged(Index, WithIndex : LongInt);
  begin
    if fOrientation = oSortByCols then
      begin
        fIntCells.Exchange(Index,   WithIndex);
        fFloatCells.Exchange(Index, WithIndex);
      end;
  end;

procedure tSortGrid.GridColsInserted(AtCol, Count : LongInt);
  var     i : LongInt;
  begin
    {$ifdef fpc}
      i := inherited ColCount;
      NewGridColCount(i - Count, i);
    {$endif}
    if fOrientation = oSortByCols then
      begin
        for i := inherited ColCount - 1 downto AtCol + Count do
          begin
            fAlignment[i]    := fAlignment[i - Count];
            fSortOrder[i]    := fSortOrder[i - Count];
            fFormatType[i]   := fFormatType[i - Count];
            fFormatString[i] := fFormatString[i - Count];
          end;
        for i := AtCol to AtCol + Count - 1 do
          begin
            fAlignment[i]    := taLeftJustify;
            fSortOrder[i]    := soAscending;
            fFormatType[i]   := sgfText;
            fFormatString[i] := '';
          end;
        for i := 0 to fSortColRows.Count - 1 do
          if fSortColRows[i] >= AtCol then
            fSortColRows[i] := fSortColRows[i] + Count;
      end;
  end;

procedure tSortGrid.GridRowsInserted(AtRow, Count : LongInt);
  var     i : LongInt;
  begin
    {$ifdef fpc}
      i := inherited RowCount;
      NewGridRowCount(i - Count, i);
    {$endif}
    if fOrientation = oSortByRows then
      begin
        for i := inherited RowCount - 1 downto AtRow + Count do
          begin
            fAlignment[i]    := fAlignment[i - Count];
            fSortOrder[i]    := fSortOrder[i - Count];
            fFormatType[i]   := fFormatType[i - Count];
            fFormatString[i] := fFormatString[i - Count];
          end;
        for i := AtRow to AtRow + Count - 1 do
          begin
            fAlignment[i]    := taLeftJustify;
            fSortOrder[i]    := soAscending;
            fFormatType[i]   := sgfText;
            fFormatString[i] := '';
          end;
        for i := 0 to fSortColRows.Count - 1 do
          if fSortColRows[i] >= AtRow then
            fSortColRows[i] := fSortColRows[i] + Count;
      end;
  end;

procedure tSortGrid.GridColMoved(FromIndex, ToIndex : LongInt);
  var     i                : LongInt;
          FromAlignment    : tAlignment;
          FromSortOrder    : TSortOrder;
          FromFormatType   : tFormatType;
          FromFormatString : String;
  begin
    if fOrientation = oSortByCols then
      begin
        FromSortOrder    := fSortOrder[FromIndex];
        FromFormatType   := fFormatType[FromIndex];
        FromFormatString := fFormatString[FromIndex];
        FromAlignment    := fAlignment[FromIndex];
        if ToIndex > FromIndex then
          begin
            for i := FromIndex to ToIndex - 1 do
              begin
                fSortOrder[i]    := fSortOrder[i + 1];
                fFormatType[i]   := fFormatType[i + 1];
                fFormatString[i] := fFormatString[i + 1];
                fAlignment[i]    := fAlignment[i + 1]
              end;
            for i := 0 to fSortColRows.Count - 1 do
              if fSortColRows[i] = FromIndex then
                fSortColRows[i] := ToIndex
              else
                if (FromIndex < fSortColRows[i])
                   and (fSortColRows[i] <= ToIndex)
                then
                  fSortColRows[i] := fSortColRows[i] - 1;
          end
        else
          begin
            for i := FromIndex downto ToIndex + 1  do
              begin
                fSortOrder[i]    := fSortOrder[i - 1];
                fFormatType[i]   := fFormatType[i - 1];
                fFormatString[i] := fFormatString[i - 1];
                fAlignment[i]    := fAlignment[i - 1];
              end;
            for i := 0 to fSortColRows.Count - 1 do
              if fSortColRows[i] = FromIndex then
                fSortColRows[i] := ToIndex
              else
                if (ToIndex <= fSortColRows[i])
                   and (fSortColRows[i] < FromIndex)
                then
                  fSortColRows[i] := fSortColRows[i] + 1;
          end;
        fSortOrder[ToIndex]    := FromSortOrder;
        fFormatType[ToIndex]   := FromFormatType;
        fFormatString[ToIndex] := FromFormatString;
        fAlignment[ToIndex]    := FromAlignment;
      end;
  end;

procedure tSortGrid.GridRowMoved(FromIndex, ToIndex : LongInt);
  var     i                : LongInt;
          FromAlignment    : tAlignment;
          FromSortOrder    : TSortOrder;
          FromFormatType   : tFormatType;
          FromFormatString : String;
  begin
    if fOrientation = oSortByRows then
      begin
        FromSortOrder    := fSortOrder[FromIndex];
        FromFormatType   := fFormatType[FromIndex];
        FromFormatString := fFormatString[FromIndex];
        FromAlignment    := fAlignment[FromIndex];
        if ToIndex > FromIndex then
          begin
            for i := FromIndex to ToIndex - 1 do
              begin
                fSortOrder[i]    := fSortOrder[i + 1];
                fFormatType[i]   := fFormatType[i + 1];
                fFormatString[i] := fFormatString[i + 1];
                fAlignment[i]    := fAlignment[i + 1]
              end;
            for i := 0 to fSortColRows.Count - 1 do
              if fSortColRows[i] = FromIndex then
                fSortColRows[i] := ToIndex
              else
                if (FromIndex < fSortColRows[i])
                   and (fSortColRows[i] <= ToIndex)
                then
                  fSortColRows[i] := fSortColRows[i] - 1;
          end
        else
          begin
            for i := FromIndex downto ToIndex + 1  do
              begin
                fSortOrder[i]    := fSortOrder[i - 1];
                fFormatType[i]   := fFormatType[i - 1];
                fFormatString[i] := fFormatString[i - 1];
                fAlignment[i]    := fAlignment[i - 1];
              end;
            for i := 0 to fSortColRows.Count - 1 do
              if fSortColRows[i] = FromIndex then
                fSortColRows[i] := ToIndex
              else
                if (ToIndex <= fSortColRows[i])
                   and (fSortColRows[i] < FromIndex)
                then
                  fSortColRows[i] := fSortColRows[i] + 1;
          end;
        fSortOrder[ToIndex]    := FromSortOrder;
        fFormatType[ToIndex]   := FromFormatType;
        fFormatString[ToIndex] := FromFormatString;
        fAlignment[ToIndex]    := FromAlignment;
      end;
  end;

{$ifdef fpc}
  procedure tSortGrid.ColRowDeleted(IsColumn: Boolean; index: LongInt);
    begin
      inherited ColRowDeleted(IsColumn, index);
      if IsColumn then
        GridColsDeleted(index, 1)
      else
        GridRowsDeleted(index, 1);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.ColRowDeleted while sorting')
    end;

  procedure tSortGrid.ColRowExchanged(IsColumn: Boolean;
                                      index, WithIndex: LongInt);
    begin
      inherited ColRowExchanged(IsColumn, index, WithIndex);
      if IsColumn then
        GridColsExchanged(index, WithIndex)
      else
        GridRowsExchanged(index, WithIndex);
    end;

  procedure tSortGrid.ColRowInserted(IsColumn: boolean; index: LongInt);
    begin
      inherited ColRowInserted(IsColumn, index);
      if IsColumn then
        GridColsInserted(Index, 1)
      else
        GridRowsInserted(Index, 1);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.ColRowInserted while sorting');
    end;

  procedure tSortGrid.ColRowMoved(IsColumn: Boolean;
                                  FromIndex, ToIndex: LongInt);
    begin
      inherited ColRowMoved(IsColumn, FromIndex, ToIndex);
      if IsColumn then
        GridColMoved(FromIndex, ToIndex)
      else
        GridRowMoved(FromIndex, ToIndex);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.GridColRowMoved while sorting');
    end;
{$else}

  procedure tSortGrid.ColRowExchanged(IsColumn : Boolean;
                                      Index, WithIndex : LongInt);
    var     i : LongInt;
    begin
      if IsColumn then
        begin
          for i := 0 to inherited RowCount - 1 do
            Rows[i].Exchange(Index, WithIndex);
          GridColsExchanged(Index, WithIndex);
        end
      else
        begin
          for i := 0 to inherited ColCount - 1 do
            Cols[i].Exchange(Index, WithIndex);
          GridRowsExchanged(Index, WithIndex);
        end;
      if Assigned(fOnColRowExchanged) then
        fOnColRowExchanged(Self, IsColumn, Index, WithIndex);
    end;

  procedure tSortGrid.ColumnMoved(FromIndex, ToIndex : LongInt);
    begin
      inherited ColumnMoved(FromIndex, ToIndex);
      GridColMoved(FromIndex, ToIndex);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.ColumnMoved while sorting');
    end;

  procedure tSortGrid.RowMoved(FromIndex, ToIndex : LongInt);
    begin
      inherited RowMoved(FromIndex, ToIndex);
      GridRowMoved(FromIndex, ToIndex);
      if fIsSorting then
        raise Exception.Create(
          'must not tSortGrid.RowMoved while sorting');
    end;
{$endif}

procedure tSortGrid.SetEditText(ACol, ARow: Longint; const Value: String);
  begin
    inherited SetEditText(ACol, ARow, Value);
    CheckSortCol(ACol);
  end;

procedure tSortGrid.DoSort(FromColRow, ToColRow : LongInt);
  var     i, j   : LongInt;
          Format,
          Error  : String;
  begin
    if fIsSorting then
      raise Exception.Create('tSortGrid.Sort: already sorting')
    else
      begin
        if fSortColRows.Count > 0 then
          begin
            fMayContinue := True;
            fIsSorting   := True;
            Error        := '';
            if FromColRow > ToColRow then
              begin
                i          := FromColRow;
                FromColRow := ToColRow;
                ToColRow   := i;
              end;
            try
              if fOrientation = oSortByCols then
                for i := 0 to Inherited ColCount - 1 do
                  case fFormatType[i] of
                    sgfInteger:
                      begin
                        fIntCells.AddInts(ToColRow + 1);
                        fFloatCells.AddNil;
                        for j := FromColRow to ToColRow do
                          if Trim(inherited Cells[i, j]) = '' then
                            fIntCells[i, j] := 0
                          else
                            try
                              fIntCells[i, j] :=
                                DezToInt(inherited Cells[i, j]);
                            except
                              fIntCells[i, j] := 0;
                            end;
                      end;
                    sgfFloat:
                      begin
                        fFloatCells.AddExtended(ToColRow + 1);
                        fIntCells.AddNil;
                        for j := FromColRow to ToColRow do
                          if Trim(inherited Cells[i, j]) = '' then
                            fFloatCells[i, j] := 0
                          else
                            try
                              fFloatCells[i, j] :=
                                StrToFloat(StringReplace(inherited Cells[i, j],
                                                         GetThousandSeparator,
                                                         '', [rfReplaceAll]));
                            except
                              fFloatCells[i, j] := 0;
                            end;
                      end;
                    sgfDateTime:
                      begin
                        fFloatCells.AddExtended(ToColRow + 1);
                        fIntCells.AddNil;
                        Format := Trim(FormatString[i]);
                        if Format = '' then
                          for j := FromColRow to ToColRow do
                            if Trim(inherited Cells[i, j]) = '' then
                              fFloatCells[i, j] := 0
                            else
                              try
                                fFloatCells[i, j] :=
                                  StrToDateTime(inherited Cells[i, j])
                              except
                                fFloatCells[i, j] := 0;
                              end
                        else
                          for j := FromColRow to ToColRow do
                            if Trim(inherited Cells[i, j]) = '' then
                              fFloatCells[i, j] := 0
                            else
                              try
                                fFloatCells[i, j] :=
                                  StrWithFormatToDateTime(
                                    inherited Cells[i, j], Format);
                              except
                                fFloatCells[i, j] := 0;
                              end;
                      end;
                  else
                    begin
                      fIntCells.AddNil;
                      fFloatCells.AddNil;
                    end;
                  end
              else
                for i := 0 to Inherited RowCount - 1 do
                  case fFormatType[i] of
                    sgfInteger:
                      begin
                        fIntCells.AddInts(ToColRow + 1);
                        fFloatCells.AddNil;
                        for j := FromColRow to ToColRow do
                          if Trim(inherited Cells[j, i]) = '' then
                            fIntCells[i, j] := 0
                          else
                            try
                              fIntCells[i, j] :=
                                DezToInt(inherited Cells[j, i]);
                            except
                              fIntCells[i, j] := 0;
                            end;
                      end;
                    sgfFloat:
                      begin
                        fFloatCells.AddExtended(ToColRow + 1);
                        fIntCells.AddNil;
                        for j := FromColRow to ToColRow do
                          if Trim(inherited Cells[j, i]) = '' then
                            fFloatCells[i, j] := 0
                          else
                            try
                              fFloatCells[i, j] :=
                                StrToFloat(StringReplace(inherited Cells[j, i],
                                                         GetThousandSeparator,
                                                         '', [rfReplaceAll]));
                            except
                              fFloatCells[i, j] := 0;
                            end;
                      end;
                    sgfDateTime:
                      begin
                        fFloatCells.AddExtended(ToColRow + 1);
                        fIntCells.AddNil;
                        Format := Trim(fFormatString[i]);
                        if Format = '' then
                          for j := FromColRow to ToColRow do
                            if Trim(inherited Cells[j, i]) = '' then
                              fFloatCells[i, j] := 0
                            else
                              try
                                fFloatCells[i, j] :=
                                  StrToDateTime(inherited Cells[j, i])
                              except
                                fFloatCells[i, j] := 0;
                              end
                        else
                          for j := FromColRow to ToColRow do
                            if Trim(inherited Cells[j, i]) = '' then
                              fFloatCells[i, j] := 0
                            else
                              try
                                fFloatCells[i, j] :=
                                  StrWithFormatToDateTime(
                                    inherited Cells[j, i], Format);
                              except
                                fFloatCells[i, j] := 0;
                              end;
                      end;
                  else
                    begin
                      fIntCells.AddNil;
                      fFloatCells.AddNil;
                    end;
                  end;
            except
              on e : Exception do
                Error := e.Message;
            end;
            if Error > '' then
              raise Exception.Create(Error)
            else
              fSort.Sort(FromColRow, ToColRow);
            fIntCells.Clear;
            fFloatCells.Clear;
            fIsSorting := False;
            if fMayContinue then
              fIsSorted := True;
            if fOrientation = oSortByCols then
              InvalidateCell(fSortColRows[0], 0)
            else
              InvalidateCell(0, fSortColRows[0]);
          end;
      end;
  end;

procedure tSortGrid.DeleteRows(FromRow, ToRow : LongInt);
  var     i, Count : LongInt;
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.DeleteRows while sorting')
    else
      begin
        if ToRow < FromRow then
          begin
            i       := ToRow;
            ToRow   := FromRow;
            FromRow := i;
          end;
        Count := ToRow - FromRow + 1;
        {$ifdef fpc}
          for i := 1 to Count do
            DeleteRow(FromRow);
        {$else}
          for i := FromRow to inherited RowCount - 1 - Count do
            Rows[i] := Rows[i + Count];
          GridRowsDeleted(FromRow, Count);
          inherited RowCount := inherited RowCount - Count;
        {$endif}
      end;
  end;

procedure tSortGrid.InsertRows(AtRow, Count : LongInt);
  var     i        : LongInt;
          {$ifdef fpc}
            r      : LongInt;
          {$endif}
  begin
    if fIsSorting then
      raise Exception.Create('must not tSortGrid.InsertRows while sorting')
    else
      begin
        {$ifdef fpc}
          r := Row;
          for i := 1 to count do
            InsertColRow(False, AtRow);
          if r = AtRow then
            Row := r;
        {$else}
          SetGridRowCount(inherited RowCount + Count);
          for i := inherited RowCount - 1 downto AtRow + Count do
            Rows[i]      := Rows[i - Count];
          for i := AtRow to AtRow + Count - 1 do
            Rows[i].Clear;
          GridRowsInserted(AtRow, Count);
        {$endif}
      end;
  end;

procedure tSortGrid.InsertCols(AtCol, Count : LongInt);
  var     i : LongInt;
          {$ifdef fpc}
            c      : LongInt;
          {$endif}
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.InsertCols while sorting')
    else
      begin
        {$ifdef fpc}
          c := Col;
          for i := 1 to Count do
            InsertColRow(True, AtCol);
          if c = AtCol then
            Col := c;
        {$else}
          SetGridColCount(inherited ColCount + Count);
          for i := inherited ColCount - 1 downto AtCol + Count do
            begin
              Cols[i]      := Cols[i - Count];
              ColWidths[i] := ColWidths[i - Count];
            end;
          for i := AtCol to AtCol + Count - 1 do
            Cols[i].Clear;
          GridColsInserted(AtCol, Count);
        {$endif}
      end;
  end;

procedure tSortGrid.DeleteCols(FromCol, ToCol : LongInt);
  var     i, Count : LongInt;
  begin
    if fIsSorting then
      raise Exception.Create(
        'must not tSortGrid.DeleteCols while sorting')
    else
      begin
        if ToCol < FromCol then
          begin
            i       := ToCol;
            ToCol   := FromCol;
            FromCol := i;                
          end;
        Count := ToCol - FromCol + 1;
        {$ifdef fpc}
          for i := 1 to Count do
            DeleteCol(FromCol);
        {$else}
          for i := FromCol to inherited ColCount - 1 - Count do
            begin
              Cols[i]      := Cols[i + Count];
              ColWidths[i] := ColWidths[i + Count];
            end;
          GridColsDeleted(FromCol, Count);
          SetGridColCount(inherited ColCount - Count);
        {$endif}
      end;
  end;

procedure tSortGrid.ExactWidthCell(ACol, ARow : LongInt);
  begin
    ColWidths[ACol] := Canvas.TextWidth(inherited Cells[ACol, ARow])
                       + GridLineWidth + MinWidth;
  end;

procedure tSortGrid.ExactWidthRow(ARow : LongInt);
  var i : LongInt;
  begin
    for i := 0 to inherited ColCount - 1 do
      ExactWidthCell(i, ARow);
  end;

procedure tSortGrid.MinWidthCell(ACol, ARow : LongInt);
  var w : LongInt;
  begin
    w := Canvas.TextWidth(inherited Cells[ACol, ARow])
         + GridLineWidth + MinWidth;
    if ColWidths[ACol] < w then
      ColWidths[ACol] := w;
  end;

procedure tSortGrid.MinWidthRow(ARow : LongInt);
  var i : LongInt;
  begin
    for i := 0 to inherited ColCount - 1 do
      MinWidthCell(i, ARow);
  end;

function tSortGrid.RequiredWidthCol(ACol : LongInt) : LongInt;
  var    i, w, c : LongInt;
  begin
    c := 0;
    for i := 0 to inherited RowCount - 1 do
      begin
        w := Canvas.TextWidth(inherited Cells[ACol, i])
             + GridLineWidth + MinWidth;
        if w > c then
          c := w;
      end;
    RequiredWidthCol := c;
  end;

procedure tSortGrid.MinWidthCol(ACol : LongInt);
  var w : LongInt;
  begin
    w := RequiredWidthCol(ACol);
    if ColWidths[ACol] < w then
      ColWidths[ACol] := w;
  end;

procedure tSortGrid.ExactWidthCol(ACol : LongInt);
  begin
    ColWidths[ACol] := RequiredWidthCol(ACol);
  end;

procedure tSortGrid.FillWidth(ACol : LongInt);
  var     w, i : Integer;
  begin
    w := ClientWidth;
    for i := 0 to ColCount - 1 do
      if i <> ACol then
        dec(w, ColWidths[i] + GridLineWidth);
    if ColWidths[ACol] < w then
      ColWidths[ACol] := w;
  end;

begin
  {$ifdef fpc}
    OSVERSIONINFO.dwMajorVersion := GetVersion div 100;
  {$else}
    OSVERSIONINFO.dwOSVersionInfoSize := sizeof(OSVERSIONINFO);
    if not GetVersionEx(OSVERSIONINFO) then
      FillChar(OSVERSIONINFO, 0, sizeof(OSVERSIONINFO));
  {$endif}
end.
